@@ -21,6 +21,7 @@ static const char *cbz_ext_list[] = {
 	".pam",
 	".pbm",
 	".pgm",
+	".pkm",
 	".png",
 	".pnm",
 	".ppm",
@@ -101,7 +102,7 @@ cbz_create_page_list(fz_context *ctx, cbz_document *doc)
 	count = fz_count_archive_entries(ctx, arch);
 
 	doc->page_count = 0;
-	doc->page = fz_malloc_array(ctx, count, sizeof *doc->page);
+	doc->page = fz_malloc_array(ctx, count, const char *);
 
 	for (i = 0; i < count; i++)
 	{
@@ -135,25 +136,26 @@ cbz_count_pages(fz_context *ctx, fz_document *doc_)
 	return doc->page_count;
 }
 
-static fz_rect *
-cbz_bound_page(fz_context *ctx, fz_page *page_, fz_rect *bbox)
+static fz_rect
+cbz_bound_page(fz_context *ctx, fz_page *page_)
 {
 	cbz_page *page = (cbz_page*)page_;
 	fz_image *image = page->image;
 	int xres, yres;
+	fz_rect bbox;
 
 	fz_image_resolution(image, &xres, &yres);
-	bbox->x0 = bbox->y0 = 0;
-	bbox->x1 = image->w * DPI / xres;
-	bbox->y1 = image->h * DPI / yres;
+	bbox.x0 = bbox.y0 = 0;
+	bbox.x1 = image->w * DPI / xres;
+	bbox.y1 = image->h * DPI / yres;
 	return bbox;
 }
 
 static void
-cbz_run_page(fz_context *ctx, fz_page *page_, fz_device *dev, const fz_matrix *ctm, fz_cookie *cookie)
+cbz_run_page(fz_context *ctx, fz_page *page_, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
 {
 	cbz_page *page = (cbz_page*)page_;
-	fz_matrix local_ctm = *ctm;
+	fz_matrix local_ctm;
 	fz_image *image = page->image;
 	int xres, yres;
 	float w, h;
@@ -161,8 +163,8 @@ cbz_run_page(fz_context *ctx, fz_page *page_, fz_device *dev, const fz_matrix *c
 	fz_image_resolution(image, &xres, &yres);
 	w = image->w * DPI / xres;
 	h = image->h * DPI / yres;
-	fz_pre_scale(&local_ctm, w, h);
-	fz_fill_image(ctx, dev, image, &local_ctm, 1, NULL);
+	local_ctm = fz_pre_scale(ctm, w, h);
+	fz_fill_image(ctx, dev, image, local_ctm, 1, fz_default_color_params);
 }
 
 static void
@@ -180,7 +182,7 @@ cbz_load_page(fz_context *ctx, fz_document *doc_, int number)
 	fz_buffer *buf = NULL;
 
 	if (number < 0 || number >= doc->page_count)
-		return NULL;
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot load page %d", number);
 
 	fz_var(page);
 
