@@ -1,3 +1,4 @@
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
 typedef struct pdf_output_processor_s pdf_output_processor;
@@ -48,9 +49,10 @@ static void
 pdf_out_d(fz_context *ctx, pdf_processor *proc, pdf_obj *array, float phase)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
+	int ahx = ((pdf_output_processor*)proc)->ahxencode;
 	if (!((pdf_output_processor*)proc)->extgstate)
 	{
-		pdf_print_obj(ctx, out, array, 1);
+		pdf_print_obj(ctx, out, array, 1, ahx);
 		fz_write_printf(ctx, out, " %g d\n", phase);
 	}
 }
@@ -60,7 +62,27 @@ pdf_out_ri(fz_context *ctx, pdf_processor *proc, const char *intent)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	if (!((pdf_output_processor*)proc)->extgstate)
-		fz_write_printf(ctx, out, "/%s ri\n", intent);
+		fz_write_printf(ctx, out, "%n ri\n", intent);
+}
+
+static void
+pdf_out_gs_OP(fz_context *ctx, pdf_processor *proc, int b)
+{
+}
+
+static void
+pdf_out_gs_op(fz_context *ctx, pdf_processor *proc, int b)
+{
+}
+
+static void
+pdf_out_gs_OPM(fz_context *ctx, pdf_processor *proc, int i)
+{
+}
+
+static void
+pdf_out_gs_UseBlackPtComp(fz_context *ctx, pdf_processor *proc, pdf_obj *name)
+{
 }
 
 static void
@@ -76,7 +98,7 @@ pdf_out_gs_begin(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	((pdf_output_processor*)proc)->extgstate = 1;
-	fz_write_printf(ctx, out, "/%s gs\n", name);
+	fz_write_printf(ctx, out, "%n gs\n", name);
 }
 
 static void
@@ -282,6 +304,7 @@ pdf_out_Tw(fz_context *ctx, pdf_processor *proc, float wordspace)
 static void
 pdf_out_Tz(fz_context *ctx, pdf_processor *proc, float scale)
 {
+	/* scale is exactly as read from the file. */
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	fz_write_printf(ctx, out, "%g Tz\n", scale);
 }
@@ -298,7 +321,9 @@ pdf_out_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_font_desc
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	if (!((pdf_output_processor*)proc)->extgstate)
-		fz_write_printf(ctx, out, "/%s %g Tf\n", name, size);
+	{
+		fz_write_printf(ctx, out, "%n %g Tf\n", name, size);
+	}
 }
 
 static void
@@ -385,7 +410,8 @@ static void
 pdf_out_TJ(fz_context *ctx, pdf_processor *proc, pdf_obj *array)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	pdf_print_obj(ctx, out, array, 1);
+	int ahx = ((pdf_output_processor*)proc)->ahxencode;
+	pdf_print_obj(ctx, out, array, 1, ahx);
 	fz_write_string(ctx, out, " TJ\n");
 }
 
@@ -436,14 +462,14 @@ static void
 pdf_out_CS(fz_context *ctx, pdf_processor *proc, const char *name, fz_colorspace *cs)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s CS\n", name);
+	fz_write_printf(ctx, out, "%n CS\n", name);
 }
 
 static void
 pdf_out_cs(fz_context *ctx, pdf_processor *proc, const char *name, fz_colorspace *cs)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s cs\n", name);
+	fz_write_printf(ctx, out, "%n cs\n", name);
 }
 
 static void
@@ -453,7 +479,7 @@ pdf_out_SC_pattern(fz_context *ctx, pdf_processor *proc, const char *name, pdf_p
 	int i;
 	for (i = 0; i < n; ++i)
 		fz_write_printf(ctx, out, "%g ", color[i]);
-	fz_write_printf(ctx, out, "/%s SCN\n", name);
+	fz_write_printf(ctx, out, "%n SCN\n", name);
 }
 
 static void
@@ -463,21 +489,21 @@ pdf_out_sc_pattern(fz_context *ctx, pdf_processor *proc, const char *name, pdf_p
 	int i;
 	for (i = 0; i < n; ++i)
 		fz_write_printf(ctx, out, "%g ", color[i]);
-	fz_write_printf(ctx, out, "/%s scn\n", name);
+	fz_write_printf(ctx, out, "%n scn\n", name);
 }
 
 static void
 pdf_out_SC_shade(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *shade)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s SCN\n", name);
+	fz_write_printf(ctx, out, "%n SCN\n", name);
 }
 
 static void
 pdf_out_sc_shade(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *shade)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s scn\n", name);
+	fz_write_printf(ctx, out, "%n scn\n", name);
 }
 
 static void
@@ -545,7 +571,7 @@ pdf_out_k(fz_context *ctx, pdf_processor *proc, float c, float m, float y, float
 /* shadings, images, xobjects */
 
 static void
-pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img)
+pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img, const char *colorspace)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	int ahx = ((pdf_output_processor*)proc)->ahxencode;
@@ -576,8 +602,10 @@ pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img)
 		fz_write_string(ctx, out, "/CS/RGB\n");
 	else if (img->colorspace == fz_device_cmyk(ctx))
 		fz_write_string(ctx, out, "/CS/CMYK\n");
-	else if (fz_colorspace_is_indexed(ctx, img->colorspace))
-		fz_write_string(ctx, out, "/CS/I\n");
+	else if (colorspace)
+		fz_write_printf(ctx, out, "/CS%n/n", colorspace);
+	else
+		fz_throw(ctx, FZ_ERROR_GENERIC, "BI operator can only show ImageMask, Gray, RGB, or CMYK images");
 	if (img->interpolate)
 		fz_write_string(ctx, out, "/I true\n");
 	fz_write_string(ctx, out, "/D[");
@@ -693,21 +721,21 @@ static void
 pdf_out_sh(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *shade)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s sh\n", name);
+	fz_write_printf(ctx, out, "%n sh\n", name);
 }
 
 static void
 pdf_out_Do_image(fz_context *ctx, pdf_processor *proc, const char *name, fz_image *image)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s Do\n", name);
+	fz_write_printf(ctx, out, "%n Do\n", name);
 }
 
 static void
-pdf_out_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_xobject *xobj, pdf_obj *page_resources)
+pdf_out_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *xobj, pdf_obj *page_resources)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s Do\n", name);
+	fz_write_printf(ctx, out, "%n Do\n", name);
 }
 
 /* marked content */
@@ -716,15 +744,16 @@ static void
 pdf_out_MP(fz_context *ctx, pdf_processor *proc, const char *tag)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s MP\n", tag);
+	fz_write_printf(ctx, out, "%n MP\n", tag);
 }
 
 static void
 pdf_out_DP(fz_context *ctx, pdf_processor *proc, const char *tag, pdf_obj *raw, pdf_obj *cooked)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s ", tag);
-	pdf_print_obj(ctx, out, raw, 1);
+	int ahx = ((pdf_output_processor*)proc)->ahxencode;
+	fz_write_printf(ctx, out, "%n ", tag);
+	pdf_print_obj(ctx, out, raw, 1, ahx);
 	fz_write_string(ctx, out, " DP\n");
 }
 
@@ -732,15 +761,16 @@ static void
 pdf_out_BMC(fz_context *ctx, pdf_processor *proc, const char *tag)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s BMC\n", tag);
+	fz_write_printf(ctx, out, "%n BMC\n", tag);
 }
 
 static void
 pdf_out_BDC(fz_context *ctx, pdf_processor *proc, const char *tag, pdf_obj *raw, pdf_obj *cooked)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_write_printf(ctx, out, "/%s ", tag);
-	pdf_print_obj(ctx, out, raw, 1);
+	int ahx = ((pdf_output_processor*)proc)->ahxencode;
+	fz_write_printf(ctx, out, "%n ", tag);
+	pdf_print_obj(ctx, out, raw, 1, ahx);
 	fz_write_string(ctx, out, " BDC\n");
 }
 
@@ -767,6 +797,13 @@ pdf_out_EX(fz_context *ctx, pdf_processor *proc)
 	fz_write_string(ctx, out, "EX\n");
 }
 
+static void
+pdf_close_output_processor(fz_context *ctx, pdf_processor *proc)
+{
+	fz_output *out = ((pdf_output_processor*)proc)->out;
+	fz_close_output(ctx, out);
+}
+
 static void
 pdf_drop_output_processor(fz_context *ctx, pdf_processor *proc)
 {
@@ -774,11 +811,21 @@ pdf_drop_output_processor(fz_context *ctx, pdf_processor *proc)
 	fz_drop_output(ctx, out);
 }
 
-static pdf_processor *
+/*
+	Create an output processor. This
+	sends the incoming PDF operator stream to an fz_output stream.
+
+	out: The output stream to which operators will be sent.
+
+	ahxencode: If 0, then image streams will be send as binary,
+	otherwise they will be asciihexencoded.
+*/
+pdf_processor *
 pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahxencode)
 {
 	pdf_output_processor *proc = pdf_new_processor(ctx, sizeof *proc);
 	{
+		proc->super.close_processor = pdf_close_output_processor;
 		proc->super.drop_processor = pdf_drop_output_processor;
 
 		/* general graphics state */
@@ -890,6 +937,12 @@ pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahxencode)
 		/* compatibility */
 		proc->super.op_BX = pdf_out_BX;
 		proc->super.op_EX = pdf_out_EX;
+
+		/* extgstate */
+		proc->super.op_gs_OP = pdf_out_gs_OP;
+		proc->super.op_gs_op = pdf_out_gs_op;
+		proc->super.op_gs_OPM = pdf_out_gs_OPM;
+		proc->super.op_gs_UseBlackPtComp = pdf_out_gs_UseBlackPtComp;
 	}
 
 	proc->out = out;
@@ -898,10 +951,20 @@ pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahxencode)
 	return (pdf_processor*)proc;
 }
 
+/*
+	Create a buffer processor. This
+	collects the incoming PDF operator stream into an fz_buffer.
+
+	buffer: The (possibly empty) buffer to which operators will be
+	appended.
+
+	ahxencode: If 0, then image streams will be send as binary,
+	otherwise they will be asciihexencoded.
+*/
 pdf_processor *
 pdf_new_buffer_processor(fz_context *ctx, fz_buffer *buffer, int ahxencode)
 {
-	pdf_processor *proc;
+	pdf_processor *proc = NULL;
 	fz_output *out = fz_new_output_with_buffer(ctx, buffer);
 	fz_try(ctx)
 	{
