@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -19,10 +19,6 @@
 
 /* configuration header file for compiling under Microsoft Windows */
 
-/* update package version here */
-#define PACKAGE "jbig2dec"
-#define VERSION "0.14"
-
 #ifdef _MSC_VER
 
 /* VS 2012 and later have stdint.h */
